package com.bpcl.lpg.core.encryption.pgp;

import org.slf4j.LoggerFactory;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.operator.PublicKeyDataDecryptorFactory;
import java.util.Iterator;
import org.bouncycastle.openpgp.PGPSignatureList;
import org.bouncycastle.openpgp.operator.PGPContentVerifierBuilderProvider;
import org.bouncycastle.openpgp.operator.bc.BcPGPContentVerifierBuilderProvider;
import org.bouncycastle.openpgp.PGPOnePassSignatureList;
import org.bouncycastle.openpgp.PGPOnePassSignature;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPCompressedData;
import org.bouncycastle.openpgp.operator.bc.BcPublicKeyDataDecryptorFactory;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPUtil;
import org.bouncycastle.openpgp.operator.KeyFingerPrintCalculator;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.operator.bc.BcKeyFingerprintCalculator;
import org.bouncycastle.bcpg.ArmoredInputStream;
import org.bouncycastle.openpgp.operator.PGPKeyEncryptionMethodGenerator;
import org.bouncycastle.openpgp.operator.bc.BcPublicKeyKeyEncryptionMethodGenerator;
import org.bouncycastle.openpgp.operator.PGPDataEncryptorBuilder;
import org.bouncycastle.openpgp.PGPEncryptedDataGenerator;
import org.bouncycastle.openpgp.operator.bc.BcPGPDataEncryptorBuilder;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPPrivateKey;
import java.io.IOException;
import me.sniggle.pgp.crypt.internal.io.IOUtils;
import java.util.Date;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.operator.PGPContentSignerBuilder;
import org.bouncycastle.openpgp.PGPSignatureGenerator;
import org.bouncycastle.openpgp.operator.bc.BcPGPContentSignerBuilder;
import org.bouncycastle.bcpg.BCPGOutputStream;
import org.bouncycastle.openpgp.PGPCompressedDataGenerator;
import java.io.OutputStream;
import java.io.InputStream;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.slf4j.Logger;
import me.sniggle.pgp.crypt.MessageEncryptor;
import me.sniggle.pgp.crypt.internal.BasePGPCommon;

public class PGPMessageEncryptor extends BasePGPCommon implements MessageEncryptor
{
    private static final Logger LOGGER;
    
    private void encryptAndSign(final PGPSecretKey pgpSecretKey, final String password, final String inputDataName, final InputStream inputData, final OutputStream encryptedDataStream) throws PGPException, IOException {
        PGPMessageEncryptor.LOGGER.trace("encryptAndSign(PGPSecretKey, String, String, InputStream, OutputStream)");
        PGPMessageEncryptor.LOGGER.trace("Secret Key: {}, Password: {}, Input Name: {}, Input Data: {}, Output Data: {}", (pgpSecretKey == null) ? "not set" : "set", (password == null) ? "not set" : "********", inputDataName, (inputData == null) ? "not set" : "set", (encryptedDataStream == null) ? "not set" : "set");
        PGPSignatureGenerator pgpSignatureGenerator = null;
        PGPPrivateKey signingKey = null;
        if (pgpSecretKey != null) {
            PGPMessageEncryptor.LOGGER.info("Retrieving signing key from secret key");
            signingKey = this.findPrivateKey(pgpSecretKey, password);
        }
        PGPMessageEncryptor.LOGGER.debug("Wrapping target stream in compressed data stream");
        final PGPCompressedDataGenerator compressedDataGenerator = new PGPCompressedDataGenerator(this.getCompressionAlgorithm());
        try (final OutputStream compressedDataStream = (OutputStream)new BCPGOutputStream(compressedDataGenerator.open(encryptedDataStream))) {
            if (signingKey != null) {
                PGPMessageEncryptor.LOGGER.info("Preparing message signing");
                pgpSignatureGenerator = new PGPSignatureGenerator((PGPContentSignerBuilder)new BcPGPContentSignerBuilder(signingKey.getPublicKeyPacket().getAlgorithm(), 8));
                pgpSignatureGenerator.init(0, signingKey);
                pgpSignatureGenerator.generateOnePassVersion(false).encode(compressedDataStream);
            }
            else {
                PGPMessageEncryptor.LOGGER.info("No signing key provided. Encrypted data will be unsigned!");
            }
            final PGPLiteralDataGenerator literalDataGenerator = new PGPLiteralDataGenerator(false);
            PGPMessageEncryptor.LOGGER.debug("Wrapping compressed data stream in literal data stream");
            try (final OutputStream literalDataOutputStream = literalDataGenerator.open(compressedDataStream, 'b', inputDataName, new Date(), new byte[4096])) {
                IOUtils.StreamHandler streamHandler = null;
                if (signingKey != null) {
                    final PGPSignatureGenerator callbackGenerator = pgpSignatureGenerator;
                    streamHandler = (IOUtils.StreamHandler)new IOUtils.StreamHandler() {
                        public void handleStreamBuffer(final byte[] buffer, final int offset, final int length) throws IOException {
                            callbackGenerator.update(buffer, offset, length);
                        }
                    };
                }
                PGPMessageEncryptor.LOGGER.info("Encrypting data and saving to target stream");
                IOUtils.copy(inputData, literalDataOutputStream, new byte[4096], streamHandler);
                literalDataGenerator.close();
            }
            if (signingKey != null) {
                PGPMessageEncryptor.LOGGER.info("Generating data signature");
                pgpSignatureGenerator.generate().encode(compressedDataStream);
            }
            compressedDataStream.close();
        }
    }
    
    protected int getEncryptionAlgorithm() {
        return this.isUnlimitedEncryptionStrength() ? 9 : 7;
    }
    
    public boolean encrypt(final InputStream publicKeyOfRecipient, final String inputDataName, final InputStream plainInputData, final OutputStream target) {
        PGPMessageEncryptor.LOGGER.trace("encrypt(InputStream, String, InputStream, OutputStream)");
        PGPMessageEncryptor.LOGGER.trace("Public Key: {}, Input Name: {}, Input Data: {}, Output: {}", (publicKeyOfRecipient == null) ? "not set" : "set", inputDataName, (plainInputData == null) ? "not set" : "set", (target == null) ? "not set" : "set");
        return this.encrypt(publicKeyOfRecipient, null, null, null, inputDataName, plainInputData, target);
    }
    
    public boolean encrypt(final InputStream publicKeyOfRecipient, final InputStream privateKeyOfSender, final String userIdOfSender, final String passwordOfSendersPrivateKey, final String inputDataName, final InputStream plainInputData, final OutputStream target) {
        PGPMessageEncryptor.LOGGER.trace("encrypt(InputStream, InputStream, String, String, String, InputStream, OutputStream)");
        PGPMessageEncryptor.LOGGER.trace("Public Key: {}, Private Key: {}, User ID: {}, Password: {}, Input Name: {}, Input Data: {}, Output: {}", (publicKeyOfRecipient == null) ? "not set" : "set", (privateKeyOfSender == null) ? "not set" : "set", userIdOfSender, (passwordOfSendersPrivateKey == null) ? "not set" : "********", inputDataName, (plainInputData == null) ? "not set" : "set", (target == null) ? "not set" : "set");
        boolean result = true;
        PGPMessageEncryptor.LOGGER.debug("Reading public key");
        final PGPPublicKey pgpPublicKey = this.findPublicKey(publicKeyOfRecipient, (BasePGPCommon.KeyFilter)new BasePGPCommon.KeyFilter<PGPPublicKey>() {
            public boolean accept(final PGPPublicKey pgpKey) {
                return pgpKey.isEncryptionKey() && !pgpKey.isMasterKey();
            }
        });
        if (pgpPublicKey != null) {
            PGPMessageEncryptor.LOGGER.debug("Wrapping target stream in ArmoredOutputStream");
            try (final OutputStream wrappedTargetStream = (OutputStream)new ArmoredOutputStream(target)) {
                final BcPGPDataEncryptorBuilder encryptorBuilder = new BcPGPDataEncryptorBuilder(this.getEncryptionAlgorithm());
                PGPMessageEncryptor.LOGGER.debug("Enabling integrity packet");
                encryptorBuilder.setWithIntegrityPacket(true);
                PGPMessageEncryptor.LOGGER.debug("Creating encrypted data generator");
                final PGPEncryptedDataGenerator encryptedDataGenerator = new PGPEncryptedDataGenerator((PGPDataEncryptorBuilder)encryptorBuilder);
                encryptedDataGenerator.addMethod((PGPKeyEncryptionMethodGenerator)new BcPublicKeyKeyEncryptionMethodGenerator(pgpPublicKey));
                PGPSecretKey pgpSecretKey = null;
                if (privateKeyOfSender != null) {
                    PGPMessageEncryptor.LOGGER.debug("Looking up secret key");
                    pgpSecretKey = this.findSecretKey(privateKeyOfSender, userIdOfSender);
                }
                else {
                    PGPMessageEncryptor.LOGGER.info("No private key provided -> No signing of encrypted data");
                }
                PGPMessageEncryptor.LOGGER.debug("Wrapping target stream in encrypted output stream");
                try (final OutputStream encryptedDataStream = encryptedDataGenerator.open(wrappedTargetStream, new byte[4096])) {
                    PGPMessageEncryptor.LOGGER.info("Encrypting and optionally signing of input data");
                    this.encryptAndSign(pgpSecretKey, passwordOfSendersPrivateKey, inputDataName, plainInputData, encryptedDataStream);
                }
            }
            catch (final IOException | PGPException e) {
                PGPMessageEncryptor.LOGGER.error("{}", e.getMessage());
                result &= false;
            }
        }
        else {
            PGPMessageEncryptor.LOGGER.error("No public key found for encryption!");
            result &= false;
        }
        return result;
    }
    
    public boolean decrypt(final String passwordOfReceiversPrivateKey, final InputStream privateKeyOfReceiver, final InputStream encryptedData, final OutputStream target) {
        PGPMessageEncryptor.LOGGER.trace("decrypt(String, InputStream, InputStream, OutputStream)");
        PGPMessageEncryptor.LOGGER.trace("Password: {}, Private Key: {}, Encrypted Data: {}, Output: {}", (passwordOfReceiversPrivateKey == null) ? "not set" : "********", (privateKeyOfReceiver == null) ? "not set" : "set", (encryptedData == null) ? "not set" : "set", (target == null) ? "not set" : "set");
        return this.decrypt(passwordOfReceiversPrivateKey, privateKeyOfReceiver, null, encryptedData, target);
    }
    
    public boolean decrypt(final String passwordOfReceiversPrivateKey, final InputStream privateKeyOfReceiver, final InputStream publicKeyOfSender, final InputStream encryptedData, final OutputStream target) {
        PGPMessageEncryptor.LOGGER.trace("decrypt(String, InputStream, InputStream, InputStream, OutputStream)");
        PGPMessageEncryptor.LOGGER.trace("Password: {}, Private Key: {}, Public Key: {}, Encrypted Data: {}, Output: {}", (passwordOfReceiversPrivateKey == null) ? "not set" : "set", (privateKeyOfReceiver == null) ? "not set" : "set", (publicKeyOfSender == null) ? "not set" : "set", (encryptedData == null) ? "not set" : "set", (target == null) ? "not set" : "set");
        boolean result = true;
        try {
            PGPPublicKeyRingCollection publicKeyRingCollection = null;
            if (publicKeyOfSender != null) {
                PGPMessageEncryptor.LOGGER.debug("Wrapping public key in ArmoredInputStream");
                try (final InputStream armoredInputStream = (InputStream)new ArmoredInputStream(publicKeyOfSender)) {
                    publicKeyRingCollection = new PGPPublicKeyRingCollection(armoredInputStream, (KeyFingerPrintCalculator)new BcKeyFingerprintCalculator());
                }
            }
            PGPMessageEncryptor.LOGGER.debug("Retrieving DecoderStream from encrypted input");
            try (final InputStream in = PGPUtil.getDecoderStream(encryptedData)) {
                PGPMessageEncryptor.LOGGER.debug("Create PGP Object factory");
                final PGPObjectFactory objectFactory = new PGPObjectFactory(in, (KeyFingerPrintCalculator)new BcKeyFingerprintCalculator());
                PGPMessageEncryptor.LOGGER.debug("Retrieve EncryptedDataList");
                final Object firstObject = objectFactory.nextObject();
                PGPEncryptedDataList dataList;
                if (firstObject instanceof final PGPEncryptedDataList list) {
                    dataList = list;
                }
                else {
                    dataList = (PGPEncryptedDataList)objectFactory.nextObject();
                }
                PGPPrivateKey pgpPrivateKey = null;
                PGPPublicKeyEncryptedData pgpEncryptedData = null;
                PGPMessageEncryptor.LOGGER.debug("Iterating over encrypted data objects");
                for (Iterator<PGPPublicKeyEncryptedData> iterator = dataList.getEncryptedDataObjects(); pgpPrivateKey == null && iterator.hasNext(); pgpPrivateKey = this.findPrivateKey(privateKeyOfReceiver, pgpEncryptedData.getKeyID(), passwordOfReceiversPrivateKey)) {
                    pgpEncryptedData = iterator.next();
                    PGPMessageEncryptor.LOGGER.debug("Looking up private key");
                }
                final PublicKeyDataDecryptorFactory publicKeyDataDecryptorFactory = (PublicKeyDataDecryptorFactory)new BcPublicKeyDataDecryptorFactory(pgpPrivateKey);
                PGPMessageEncryptor.LOGGER.debug("Retrieving data stream from encrypted data");
                try (final InputStream clearText = pgpEncryptedData.getDataStream(publicKeyDataDecryptorFactory)) {
                    PGPObjectFactory pgpObjectFactory = new PGPObjectFactory(clearText, (KeyFingerPrintCalculator)new BcKeyFingerprintCalculator());
                    PGPOnePassSignatureList onePassSignatureList = null;
                    PGPOnePassSignature onePassSignature = null;
                    PGPSignatureList signatures = null;
                    PGPPublicKey pgpPublicKey = null;
                    Object message;
                    while ((message = pgpObjectFactory.nextObject()) != null) {
                        if (message instanceof final PGPCompressedData compressedData) {
                            PGPMessageEncryptor.LOGGER.debug("Compressed data block found, creating new object factory with compressed data stream");
                            pgpObjectFactory = new PGPObjectFactory(compressedData.getDataStream(), (KeyFingerPrintCalculator)new BcKeyFingerprintCalculator());
                        }
                        if (message instanceof final PGPLiteralData literalData) {
                            PGPMessageEncryptor.LOGGER.debug("Reading literal data stream");
                            try (final InputStream literalDataStream = literalData.getInputStream()) {
                                IOUtils.StreamHandler streamHandler = null;
                                if (onePassSignature != null) {
                                    final PGPOnePassSignature callbackSignature = onePassSignature;
                                    streamHandler = (IOUtils.StreamHandler)new IOUtils.StreamHandler() {
                                        public void handleStreamBuffer(final byte[] buffer, final int offset, final int length) throws IOException {
                                            callbackSignature.update(buffer, offset, length);
                                        }
                                    };
                                }
                                IOUtils.copy(literalDataStream, target, new byte[4096], streamHandler);
                                if (literalDataStream == null) {
                                    continue;
                                }
                            }
                        }
                        else if (message instanceof final PGPOnePassSignatureList list2) {
                            onePassSignatureList = list2;
                            if (publicKeyOfSender == null) {
                                continue;
                            }
                            PGPMessageEncryptor.LOGGER.info("Public key provided -> verifying message signature");
                            onePassSignature = onePassSignatureList.get(0);
                            pgpPublicKey = publicKeyRingCollection.getPublicKey(onePassSignature.getKeyID());
                            onePassSignature.init((PGPContentVerifierBuilderProvider)new BcPGPContentVerifierBuilderProvider(), pgpPublicKey);
                        }
                        else {
                            if (!(message instanceof PGPSignatureList)) {
                                continue;
                            }
                            PGPMessageEncryptor.LOGGER.info("Signature List found for verification");
                            signatures = (PGPSignatureList)message;
                        }
                    }
                    PGPMessageEncryptor.LOGGER.debug("Iterating over signature list");
                    for (int i = 0; onePassSignatureList != null && i < onePassSignatureList.size(); ++i) {
                        if (pgpPublicKey != null && signatures != null) {
                            PGPMessageEncryptor.LOGGER.info("Verifying signatures");
                            final PGPSignature signature = signatures.get(i);
                            if (onePassSignature.verify(signature)) {
                                PGPMessageEncryptor.LOGGER.info("Signature verified");
                                String userId = null;
                                final Iterator<String> it = pgpPublicKey.getUserIDs();
                                while (it.hasNext()) {
                                    userId = it.next();
                                    PGPMessageEncryptor.LOGGER.info("Signed by {}", userId);
                                }
                            }
                            else {
                                PGPMessageEncryptor.LOGGER.warn("Signature verification failed");
                                result &= false;
                            }
                        }
                    }
                    if (pgpEncryptedData.isIntegrityProtected()) {
                        PGPMessageEncryptor.LOGGER.info("Performing integrity check on encrypted data");
                        if (pgpEncryptedData.verify()) {
                            PGPMessageEncryptor.LOGGER.info("Data integrity verified");
                        }
                        else {
                            PGPMessageEncryptor.LOGGER.warn("Data integrity verification failed");
                            result &= false;
                        }
                    }
                }
            }
        }
        catch (final Exception e) {
            PGPMessageEncryptor.LOGGER.error("{}" + e.getMessage(), (Throwable)e);
            result &= false;
        }
        return result;
    }
    
    static {
        LOGGER = LoggerFactory.getLogger(PGPMessageEncryptor.class);
    }
}
